package org.opengis.cite.iso19142;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.opengis.cite.iso19142.util.ServiceMetadataUtils;
import org.opengis.cite.iso19142.util.WFSClient;
import org.opengis.cite.iso19142.util.TestSuiteLogger;
import org.opengis.cite.iso19142.util.WFSRequest;
import org.opengis.cite.iso19142.util.XMLUtils;
import org.testng.ITestContext;
import org.testng.ITestResult;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import com.sun.jersey.api.client.ClientResponse;

/**
 * A supporting base class that provides common configuration methods and data
 * providers. The configuration methods are invoked before any that may be
 * defined in a subclass.
 */
public class BaseFixture {

    /**
     * A DOM document containing service metadata (OGC capabilities).
     */
    protected Document wfsMetadata;
    /**
     * A client component for interacting with a WFS.
     */
    protected WFSClient wfsClient;
    protected Set<ProtocolBinding> supportedBindings;
    protected List<QName> featureTypes;
    protected Map<QName, FeatureTypeInfo> featureInfo;
    /** A DOM parser. */
    protected DocumentBuilder docBuilder;
    protected static final String TNS_PREFIX = "tns";
    /** A Document representing the content of the request message. */
    protected Document reqEntity;

    /**
     * Sets up the base fixture. The service metadata document is obtained from
     * the ISuite context. The suite attribute
     * {@link SuiteAttribute#TEST_SUBJECT testSubject} should yield a DOM
     * Document node having {http://www.opengis.net/wfs/2.0}WFS_Capabilities as
     * the document element.
     * 
     * The set of implemented protocol bindings is determined from the service
     * metadata by checking the values of the following service constraints:
     * 
     * <ul>
     * <li>KVPEncoding</li>
     * <li>XMLEncoding</li>
     * <li>SOAPEncoding</li>
     * </ul>
     * 
     * @param testContext
     *            The test (set) context.
     */

    @BeforeClass(alwaysRun = true)
    @SuppressWarnings("unchecked")
    public void initBaseFixture(ITestContext testContext) {
        if (null == wfsMetadata) {
            this.wfsMetadata = (Document) testContext.getSuite().getAttribute(
                    SuiteAttribute.TEST_SUBJECT.getName());
        }
        this.wfsClient = new WFSClient(this.wfsMetadata);
        Set<ProtocolBinding> protoBindings = ServiceMetadataUtils
                .getSupportedBindings(wfsMetadata);
        if (protoBindings.isEmpty()) {
            TestSuiteLogger.log(Level.WARNING,
                    "No protocol bindings found in capabilities document.");
        }
        this.supportedBindings = protoBindings;
        this.featureTypes = ServiceMetadataUtils.getFeatureTypes(wfsMetadata);
        this.featureInfo = (Map<QName, FeatureTypeInfo>) testContext.getSuite()
                .getAttribute(SuiteAttribute.FEATURE_INFO.getName());
    }

    /**
     * Initializes the (namespace-aware) DOM parser.
     */
    @BeforeClass
    public void initParser() {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        try {
            this.docBuilder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            TestSuiteLogger
                    .log(Level.WARNING, "Failed to create DOM parser", e);
        }
    }

    /**
     * Augments the test result with supplementary attributes in the event that
     * a test method failed. The "request" attribute contains a String
     * representing the request entity (POST method) or query component (GET
     * method).
     */
    @AfterMethod
    public void addAttributesIfTestFailed(ITestResult result) {
        if (result.getStatus() != ITestResult.FAILURE) {
            return;
        }
        String request = null;
        Object[] params = result.getParameters();
        if ((params.length > 0)
                && ProtocolBinding.class.cast(params[0]).equals(
                        ProtocolBinding.GET)) {
            // Note: first parameter is always a ProtocolBinding object
            request = WFSRequest.transformEntityToKVP(new DOMSource(
                    this.reqEntity));
        } else {
            request = XMLUtils.writeNodeToString(this.reqEntity);
        }
        result.setAttribute("request", request);
    }

    /**
     * A DataProvider that supplies the set of application protocol bindings
     * supported by the SUT.
     * 
     * @return An Object[][] array containing a ProtocolBinding object in each
     *         row (first dimension).
     */
    @DataProvider(name = "protocol-binding")
    public Object[][] getProtocolBindings() {
        Object[][] bindings = new Object[supportedBindings.size()][];
        Iterator<ProtocolBinding> itr = supportedBindings.iterator();
        for (int i = 0; i < bindings.length; i++) {
            bindings[i] = new Object[] { itr.next() };
        }
        return bindings;
    }

    /**
     * A DataProvider that supplies the complete set of feature types recognized
     * by the SUT.
     * 
     * @return An Object[][] array containing a QName object in each row (first
     *         dimension).
     */
    @DataProvider(name = "feature-types")
    public Object[][] getFeatureTypes() {
        Object[][] typeNames = new Object[featureTypes.size()][];
        Iterator<QName> itr = featureTypes.iterator();
        for (int i = 0; i < typeNames.length; i++) {
            typeNames[i] = new Object[] { itr.next() };
        }
        return typeNames;
    }

    /**
     * A DataProvider that supplies feature type names for which instances
     * exist.
     * 
     * @return An Iterator over an array containing a QName object representing
     *         the qualified name of a feature type.
     */
    @DataProvider(name = "instantiated-feature-types")
    public Iterator<Object[]> getInstantiatedFeatureTypes() {
        List<Object[]> data = new ArrayList<Object[]>();
        for (QName typeName : this.featureTypes) {
            if (this.featureInfo.get(typeName).isInstantiated()) {
                data.add(new Object[] { typeName });
            }
        }
        return data.iterator();
    }

    /**
     * A DataProvider that supplies parameters specifying a supported protocol
     * binding and a feature type. The resulting set is given by the Cartesian
     * product of the sets {bindings} x {featureTypes}; its cardinality is equal
     * to the product of the cardinalities of the two input sets.
     * 
     * @return An Iterator<Object[]> over the set of (ProtocolBinding, QName)
     *         pairs.
     */
    @DataProvider(name = "all-protocols-featureTypes")
    public Iterator<Object[]> allProtocolsAndFeatureTypes() {
        List<Object[]> params = new ArrayList<Object[]>();
        for (ProtocolBinding binding : supportedBindings) {
            for (QName typeName : featureTypes) {
                params.add(new Object[] { binding, typeName });
            }
        }
        return params.iterator();
    }

    /**
     * A DataProvider that supplies parameters specifying a supported protocol
     * binding and an available feature type (at least one valid instance exists
     * in the data store). If the test method belongs to the "transaction"
     * group, the HTTP GET method binding is ignored.
     * 
     * @param method
     *            The test method about to be invoked.
     * 
     * @return An Iterator<Object[]> over the set of (ProtocolBinding, QName)
     *         pairs.
     */
    @DataProvider(name = "protocol-featureType")
    public Iterator<Object[]> protocolAndAvailableFeatureTypes(Method method) {
        if (null == this.supportedBindings) {
            throw new IllegalStateException(getClass().getName()
                    + " may not have been initialized!");
        }
        Test test = method.getAnnotation(Test.class);
        boolean isTransaction = Arrays.asList(test.groups()).contains(
                "transaction");
        List<Object[]> params = new ArrayList<Object[]>();
        for (ProtocolBinding binding : this.supportedBindings) {
            if (isTransaction && binding.equals(ProtocolBinding.GET))
                continue; // can't execute transactions using HTTP GET
            for (QName typeName : featureTypes) {
                if (!featureInfo.get(typeName).isInstantiated())
                    continue; // ignore unused feature type
                params.add(new Object[] { binding, typeName });
            }
        }
        return params.iterator();
    }

    /**
     * Extracts the body of the response message as a DOM Document node. For a
     * SOAP response this will contain the content of the first child of the
     * SOAP body element.
     * 
     * @param rsp
     *            A ClientResponse representing an HTTP response message.
     * @param binding
     *            The protocol binding used in the request.
     * @return A Document representing the response entity, or {@code null} if
     *         it could not be parsed.
     */
    protected Document extractBodyAsDocument(ClientResponse rsp,
            ProtocolBinding binding) {
        Document entity = rsp.getEntity(Document.class);
        if (binding.equals(ProtocolBinding.SOAP)) {
            Element contentNode;
            try {
                contentNode = (Element) XMLUtils.evaluateXPath(entity,
                        "//soap11:Body/*[1] | //soap:Body/*[1]", null,
                        XPathConstants.NODE);
            } catch (XPathExpressionException xpe) {
                throw new RuntimeException(xpe);
            }
            entity.replaceChild(contentNode, entity.getDocumentElement());
        }
        return entity;
    }
}
