package org.opengis.cite.iso19142.basic.filter;

import java.net.URI;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPathExpressionException;

import org.apache.xerces.xs.XSElementDeclaration;
import org.apache.xerces.xs.XSTypeDefinition;
import org.opengis.cite.geomatics.Extents;
import org.opengis.cite.geomatics.TopologicalRelationships;
import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.ErrorMessage;
import org.opengis.cite.iso19142.ErrorMessageKeys;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.cite.iso19142.util.AppSchemaUtils;
import org.opengis.cite.iso19142.util.ServiceMetadataUtils;
import org.opengis.cite.iso19142.util.WFSRequest;
import org.opengis.cite.iso19142.util.XMLUtils;
import org.opengis.referencing.operation.TransformException;
import org.testng.Assert;
import org.testng.SkipException;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.sun.jersey.api.client.ClientResponse;

/**
 * Tests the response to a GetFeature request that includes a BBOX predicate.
 * All conforming "Basic WFS" implementations must support this spatial
 * operator.
 * 
 * <h6 style="margin-bottom: 0.5em">Sources</h6>
 * <ul>
 * <li>ISO 19142:2010, Table 1: Conformance classes</li>
 * <li>ISO 19143:2010, cl. A.7: Test cases for minimum spatial filter</li>
 * </ul>
 */
@Test(groups = "basic")
public class BBOXOperatorTests extends QueryFilterFixture {

    private static final String XSLT_ENV2POLYGON = "/org/opengis/cite/iso19142/util/bbox2polygon.xsl";
    private XSTypeDefinition gmlGeomBaseType;

    /**
     * Creates an XSTypeDefinition object representing the
     * gml:AbstractGeometryType definition.
     */
    @BeforeClass
    public void createGeometryBaseType() {
        this.gmlGeomBaseType = model.getTypeDefinition("AbstractGeometryType",
                Namespaces.GML);
    }

    /**
     * [{@code Test}] Submits a GetFeature request with a non-specific BBOX
     * predicate. If no value reference is specified the predicate is applied to
     * all spatial properties. The response entity (wfs:FeatureCollection) must
     * be schema-valid and contain only instances of the requested type that
     * satisfy the spatial predicate.
     * 
     * @param binding
     *            The ProtocolBinding to use for this request.
     * @param featureType
     *            A QName representing the qualified name of some feature type.
     * 
     * @see "ISO 19143:2010, 7.8.3.2: BBOX operator"
     */
    @Test(dataProvider = "protocol-featureType")
    public void withBroadBBOXFilter(ProtocolBinding binding, QName featureType) {
        List<XSElementDeclaration> geomProps = AppSchemaUtils
                .getFeaturePropertiesByType(model, featureType, gmlGeomBaseType);
        if (geomProps.isEmpty()) {
            throw new SkipException("Feature type has no geometry properties: "
                    + featureType);
        }
        WFSRequest.appendSimpleQuery(this.reqEntity, featureType);
        Document gmlEnv = Extents.envelopeAsGML(featureInfo.get(featureType)
                .getGeoExtent());
        addBBOXPredicate(this.reqEntity, gmlEnv.getDocumentElement(), null);
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(reqEntity),
                binding, endpoint);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.OK.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity = extractBodyAsDocument(rsp, binding);
        Map<String, String> nsBindings = new HashMap<String, String>();
        nsBindings.put(Namespaces.WFS, "wfs");
        NodeList members;
        String xpath = "//wfs:member/*";
        try {
            members = XMLUtils.evaluateXPath(entity, xpath, nsBindings);
        } catch (XPathExpressionException e) {
            throw new RuntimeException(e);
        }
        Assert.assertTrue(members.getLength() > 0, ErrorMessage.format(
                ErrorMessageKeys.XPATH_RESULT, entity.getDocumentElement()
                        .getNodeName(), xpath));
        for (int i = 0; i < members.getLength(); i++) {
            ETSAssert.assertQualifiedName(members.item(i), featureType);
        }
    }

    /**
     * [{@code Test}] Submits a GetFeature request with a BBOX predicate
     * referring to a valid geometry property. The response shall contain only
     * features possessing a geometry value that spatially interacts (i.e. is
     * not disjoint) with the given envelope.
     * 
     * <h6 style="margin-bottom: 0.5em">Sources</h6>
     * <ul>
     * <li>ISO 19143:2010, cl. A.7: Test cases for minimum spatial filter</li>
     * <li>ISO 19143:2010, 7.8.3.2: BBOX operator</li>
     * </ul>
     * 
     * @param binding
     *            The ProtocolBinding to use for this request.
     * @param featureType
     *            A QName representing the qualified name of some feature type.
     * 
     */
    @Test(dataProvider = "protocol-featureType")
    public void withBBOXFilter(ProtocolBinding binding, QName featureType) {
        List<XSElementDeclaration> geomProps = AppSchemaUtils
                .getFeaturePropertiesByType(model, featureType, gmlGeomBaseType);
        if (geomProps.isEmpty()) {
            throw new SkipException("Feature type has no geometry properties: "
                    + featureType);
        }
        XSElementDeclaration geomProp = geomProps.get(0);
        Element valueRef = WFSRequest.createValueReference(geomProp);
        WFSRequest.appendSimpleQuery(this.reqEntity, featureType);
        Document gmlEnv = Extents.envelopeAsGML(featureInfo.get(featureType)
                .getGeoExtent());
        addBBOXPredicate(this.reqEntity, gmlEnv.getDocumentElement(), valueRef);
        ClientResponse rsp = wfsClient.submitRequest(reqEntity, binding);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.OK.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity = extractBodyAsDocument(rsp, binding);
        Map<String, String> nsBindings = new HashMap<String, String>();
        nsBindings.put(featureType.getNamespaceURI(), "ns1");
        String xpath;
        if (!geomProp.getNamespace().equals(featureType.getNamespaceURI())) {
            nsBindings.put(geomProp.getNamespace(), "ns2");
            xpath = String.format("//ns1:%s/ns2:%s/*[1]",
                    featureType.getLocalPart(), geomProp.getName());
        } else {
            xpath = String.format("//ns1:%s/ns1:%s/*[1]",
                    featureType.getLocalPart(), geomProp.getName());
        }
        NodeList geometryNodes;
        try {
            geometryNodes = XMLUtils.evaluateXPath(entity, xpath, nsBindings);
        } catch (XPathExpressionException e) {
            throw new RuntimeException(e);
        }
        Assert.assertTrue(geometryNodes.getLength() > 0, ErrorMessage.format(
                ErrorMessageKeys.XPATH_RESULT, entity.getDocumentElement()
                        .getNodeName(), xpath));
        Document gmlPolygon = XMLUtils.transform(new StreamSource(getClass()
                .getResourceAsStream(XSLT_ENV2POLYGON)), gmlEnv);
        for (int i = 0; i < geometryNodes.getLength(); i++) {
            Element geometry = (Element) geometryNodes.item(i);
            if (geometry.getElementsByTagNameNS(Namespaces.GML, "PolygonPatch")
                    .getLength() > 0) {
                // gml:MultiSurface with gml:Surface members
                geometry = surfaceToPolygon(geometry);
            }
            try {
                boolean intersects = TopologicalRelationships.intersects(
                        gmlPolygon.getDocumentElement(), geometry);
                Assert.assertTrue(intersects, ErrorMessage.format(
                        ErrorMessageKeys.PREDICATE_NOT_SATISFIED, "BBOX",
                        XMLUtils.writeNodeToString(gmlPolygon),
                        XMLUtils.writeNodeToString(geometry)));
            } catch (TransformException te) {
                throw new RuntimeException("Coordinate transformation failed.",
                        te);
            }
        }
    }

    /**
     * [{@code Test}] Submits a GetFeature request where the BBOX predicate
     * refers to a feature property (gml:description) that is not
     * geometry-valued. An exception is expected in response with status code
     * 400 and exception code {@code InvalidParameterValue}.
     * 
     * @param featureType
     *            A QName representing the qualified name of a feature type for
     *            which instances exist.
     * 
     * @see "ISO 19143:2010, 8.3: Exceptions"
     */
    @Test(dataProvider = "instantiated-feature-types")
    public void invalidGeometryOperand(QName featureType) {
        XSElementDeclaration gmlDesc = this.model.getElementDeclaration(
                "description", Namespaces.GML);
        Element valueRef = WFSRequest.createValueReference(gmlDesc);
        WFSRequest.appendSimpleQuery(this.reqEntity, featureType);
        Document gmlEnv = Extents.envelopeAsGML(featureInfo.get(featureType)
                .getGeoExtent());
        addBBOXPredicate(this.reqEntity, gmlEnv.getDocumentElement(), valueRef);
        ClientResponse rsp = wfsClient.submitRequest(reqEntity,
                ProtocolBinding.ANY);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.BAD_REQUEST.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity = rsp.getEntity(Document.class);
        String xpath = "//ows:Exception[@exceptionCode='InvalidParameterValue']";
        ETSAssert.assertXPath(xpath, entity, null);
    }

    /**
     * Replaces gml:Surface elements having a single gml:PolygonPatch with a
     * gml:Polygon element. Such a simplification facilitates a binding to JTS
     * geometry objects.
     * 
     * @param geometry
     *            A GML geometry collection.
     * @return A new DOM Element containing a simplified representation of the
     *         original geometry collection.
     * 
     * @see "ISO 19125-1: Geographic information -- Simple feature access --
     *      Part 1: Common architecture"
     */
    Element surfaceToPolygon(Element geometry) {
        Document result = XMLUtils.transform(new StreamSource(getClass()
                .getResourceAsStream("surface2polygon.xsl")), geometry);
        return result.getDocumentElement();
    }

    /**
     * Adds a BBOX spatial predicate to a GetFeature request entity.
     * 
     * @param request
     *            The request entity (/wfs:GetFeature).
     * @param envelope
     *            A DOM Element representing a gml:Envelope.
     * @param valueRef
     *            An Element (fes:ValueReference) that specifies the spatial
     *            property to check. If it is {@code null}, the predicate
     *            applies to all spatial properties.
     */
    void addBBOXPredicate(Document request, Element envelope, Element valueRef) {
        if (!request.getDocumentElement().getLocalName()
                .equals(WFS2.GET_FEATURE)) {
            throw new IllegalArgumentException("Not a GetFeature request: "
                    + request.getDocumentElement().getNodeName());
        }
        Element queryElem = (Element) request.getElementsByTagNameNS(
                Namespaces.WFS, WFS2.QUERY_ELEM).item(0);
        Element filter = request.createElementNS(Namespaces.FES, "Filter");
        queryElem.appendChild(filter);
        Element bbox = request.createElementNS(Namespaces.FES, "BBOX");
        filter.appendChild(bbox);
        if (null != valueRef) {
            bbox.appendChild(request.importNode(valueRef, true));
        }
        // import envelope node to avoid WRONG_DOCUMENT_ERR
        bbox.appendChild(request.importNode(envelope, true));
    }
}
