package org.opengis.cite.iso19142.locking;

import java.net.URI;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.logging.Level;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;

import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.ErrorMessage;
import org.opengis.cite.iso19142.ErrorMessageKeys;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.cite.iso19142.util.ServiceMetadataUtils;
import org.opengis.cite.iso19142.util.TestSuiteLogger;
import org.opengis.cite.iso19142.util.WFSRequest;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import com.sun.jersey.api.client.ClientResponse;

/**
 * Tests the response to a LockFeature request that attempts to lock feature
 * instances identified using one or more query expressions. A lock is active
 * until it either expires (default duration is 300 s) or it is released by a
 * subsequent transaction.
 * 
 * <h6 style="margin-bottom: 0.5em">Sources</h6>
 * <ul>
 * <li>ISO 19142:2010, cl. 12: LockFeature operation</li>
 * <li>ISO 19142:2010, cl. 12.4: Exceptions</li>
 * <li>ISO 19142:2010, cl. 15.2.3.1.1: Declaring support for locking</li>
 * </ul>
 */
@Test(groups = "locking")
public class LockFeatureTests extends LockingFixture {

    /**
     * Builds a DOM Document representing a LockFeature request entity.
     */
    @BeforeMethod
    public void buildSimpleLockFeatureRequest() {
        this.reqEntity = WFSRequest.createRequestEntity("LockFeature");
    }

    /**
     * [{@code Test}] An attempt to reset a non-existent lock should produce a
     * service exception with error code "LockHasExpired".
     * 
     * <p>
     * <strong>Note</strong>: The error code {@code InvalidLockId} would seem
     * more appropriate here but that may only appear in a Transaction response
     * according to Table 3.
     * </p>
     * 
     * @see "ISO 19142:2010, cl. 12.2.4.2: lockId parameter"
     */
    @Test
    public void resetNonexistentLock() {
        Map<String, QName> featureId = fetchRandomFeatureIdentifier(featureTypes);
        String gmlId = featureId.keySet().iterator().next();
        WFSRequest.appendStoredQuery(reqEntity, WFS2.QRY_GET_FEATURE_BY_ID,
                Collections.singletonMap("id", (Object) gmlId));
        reqEntity.getDocumentElement().setAttribute("lockId",
                "lock-does-not-exist");
        ClientResponse rsp = wfsClient.submitRequest(reqEntity,
                ProtocolBinding.ANY);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.BAD_REQUEST.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity = rsp.getEntity(Document.class);
        String xpath = "//ows:Exception[@exceptionCode = 'LockHasExpired']";
        ETSAssert.assertXPath(xpath, entity.getDocumentElement(), null);
    }

    /**
     * [{@code Test}] Submits a request to lock a feature instance; within this
     * interval an attempt to delete the instance without the correct lock
     * identifier should fail with exception code {@code MissingParameterValue}.
     * 
     * <h6 style="margin-bottom: 0.5em">Sources</h6>
     * <ul>
     * <li>ISO 19142:2010, cl. 12.2.4.3: Lock expiry parameter</li>
     * <li>ISO 19142:2010, cl. 15.2.3.1.2: lockId parameter</li>
     * </ul>
     */
    @Test
    public void lockFeatureAndAttemptDelete() {
        Map<String, QName> featureId = fetchRandomFeatureIdentifier(featureTypes);
        String gmlId = featureId.keySet().iterator().next();
        WFSRequest.appendStoredQuery(reqEntity, WFS2.QRY_GET_FEATURE_BY_ID,
                Collections.singletonMap("id", (Object) gmlId));
        reqEntity.getDocumentElement().setAttribute("expiry", "60");
        ClientResponse rsp = wfsClient.submitRequest(reqEntity,
                ProtocolBinding.ANY);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.OK.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity = rsp.getEntity(Document.class);
        ETSAssert.assertXPath("//wfs:LockFeatureResponse", entity, null);
        Element lockRsp = (Element) entity.getElementsByTagNameNS(
                Namespaces.WFS, WFS2.LOCK_FEATURE_RSP).item(0);
        String lockId = lockRsp.getAttribute("lockId");
        Assert.assertFalse(lockId.isEmpty(), ErrorMessage.format(
                ErrorMessageKeys.MISSING_INFOSET_ITEM, "@lockId"));
        locks.add(lockId);
        String xpath = String.format(
                "//wfs:FeaturesLocked/fes:ResourceId/@rid = '%s'", gmlId);
        ETSAssert.assertXPath(xpath, lockRsp, null);
        Document trxResponse = wfsClient.delete(featureId);
        String xpath2 = "//ows:Exception[@exceptionCode = 'MissingParameterValue']";
        ETSAssert.assertXPath(xpath2, trxResponse.getDocumentElement(), null);
    }

    /**
     * [{@code Test}] A feature instance may be locked by only one lock. An
     * attempt to establish another lock should fail with exception code
     * {@code CannotLockAllFeatures} if lockAction = "ALL" (the default value).
     * 
     * <h6 style="margin-bottom: 0.5em">Sources</h6>
     * <ul>
     * <li>ISO 19142:2010, Table 18: Keywords for LockFeature KVP encoding</li>
     * <li>ISO 19142:2010, cl. 12.2.5: State machine for WFS locking</li>
     * </ul>
     */
    @Test
    public void lockFeatureAlreadyLocked() {
        Map<String, QName> featureId = fetchRandomFeatureIdentifier(featureTypes);
        String gmlId = featureId.keySet().iterator().next();
        WFSRequest.appendStoredQuery(reqEntity, WFS2.QRY_GET_FEATURE_BY_ID,
                Collections.singletonMap("id", (Object) gmlId));
        reqEntity.getDocumentElement().setAttribute("expiry", "60");
        ClientResponse rsp = wfsClient.submitRequest(reqEntity,
                ProtocolBinding.ANY);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.OK.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity1 = rsp.getEntity(Document.class);
        Element lockRsp = (Element) entity1.getElementsByTagNameNS(
                Namespaces.WFS, WFS2.LOCK_FEATURE_RSP).item(0);
        locks.add(lockRsp.getAttribute("lockId"));
        // try to lock it again (without specifying lockId)
        reqEntity.getDocumentElement().setAttribute("expiry", "180");
        rsp = wfsClient.submitRequest(reqEntity, ProtocolBinding.ANY);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.BAD_REQUEST.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity = rsp.getEntity(Document.class);
        String xpath = "//ows:Exception[@exceptionCode = 'CannotLockAllFeatures']";
        ETSAssert.assertXPath(xpath, entity.getDocumentElement(), null);
    }

    /**
     * [{@code Test}] Locks all feature instances of a given type using default
     * values for all locking options. If a feature type has no instances it is
     * ignored. The response entity must include a lockId attribute and the
     * wfs:FeaturesLocked element. The wfs:FeatureNotLocked element must not be
     * present.
     * 
     * <h6 style="margin-bottom: 0.5em">Sources</h6>
     * <ul>
     * <li>ISO 19142:2010, cl. 12.3.2: XML-encoding</li>
     * </ul>
     * 
     * @param binding
     *            The ProtocolBinding to use for the request.
     * @param featureType
     *            A QName object denoting the feature type name.
     */
    @Test(dataProvider = "protocol-featureType")
    public void lockAllFeaturesByType(ProtocolBinding binding, QName featureType) {
        WFSRequest.appendStoredQuery(reqEntity, WFS2.QRY_GET_FEATURE_BY_TYPE,
                Collections.singletonMap("typeName", (Object) featureType));
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.LOCK_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(reqEntity),
                binding, endpoint);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.OK.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity = extractBodyAsDocument(rsp, binding);
        Element lockRsp = entity.getDocumentElement();
        Assert.assertEquals(lockRsp.getLocalName(), WFS2.LOCK_FEATURE_RSP,
                ErrorMessage.get(ErrorMessageKeys.LOCAL_NAME));
        String lockId = lockRsp.getAttribute("lockId");
        Assert.assertFalse(lockId.isEmpty(), ErrorMessage.format(
                ErrorMessageKeys.MISSING_INFOSET_ITEM,
                "@lockId in " + lockRsp.getNodeName()));
        locks.add(lockId);
        ETSAssert.assertXPath("//wfs:FeaturesLocked", lockRsp, null);
        ETSAssert.assertXPath("not(//wfs:FeaturesNotLocked)", lockRsp, null);
    }

    /**
     * Obtains the system-assigned identifier of an instance of some randomly
     * selected feature type.
     * 
     * @param featureTypes
     *            A list of QName objects representing feature type names.
     * @return A Map containing a single entry that associates the (String)
     *         value of the gml:id attribute with the feature type name (QName).
     */
    Map<String, QName> fetchRandomFeatureIdentifier(List<QName> featureTypes) {
        QName featureType = randomFeatureType(featureTypes);
        Document doc = wfsClient.getFeatureByType(featureType, 1);
        NodeList features = doc.getElementsByTagNameNS(
                featureType.getNamespaceURI(), featureType.getLocalPart());
        Map<String, QName> featureId;
        if (features.getLength() > 0) {
            Element feature = (Element) features.item(0);
            featureId = new HashMap<String, QName>();
            featureId.put(feature.getAttributeNS(Namespaces.GML, "id"),
                    featureType);
        } else {
            featureId = fetchRandomFeatureIdentifier(featureTypes);
        }
        if (TestSuiteLogger.isLoggable(Level.FINER)) {
            TestSuiteLogger.log(Level.FINER, featureId.toString());
        }
        return featureId;
    }

    /**
     * Randomly selects a feature type name from the given list.
     * 
     * @param featureTypes
     *            A list of QName objects representing feature type names.
     * @return A QName object from the list.
     */
    static QName randomFeatureType(List<QName> featureTypes) {
        Random randomIndex = new Random();
        return featureTypes.get(randomIndex.nextInt(featureTypes.size()));
    }
}
