package org.opengis.cite.iso19142.simple;

import com.sun.jersey.api.client.ClientResponse;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.Schema;
import javax.xml.xpath.XPathExpressionException;
import org.opengis.cite.iso19142.ErrorMessage;
import org.opengis.cite.iso19142.ErrorMessageKeys;
import org.opengis.cite.iso19142.BaseFixture;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.SuiteAttribute;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.util.XMLUtils;
import org.opengis.cite.iso19142.util.ServiceMetadataUtils;
import org.opengis.cite.iso19142.util.ValidationUtils;
import org.opengis.cite.iso19142.util.WFSRequest;
import org.opengis.cite.validation.SchematronValidator;
import org.testng.Assert;
import org.testng.ITestContext;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Tests the service response to a GetFeature request that invokes a stored
 * query. A WFS implementation is required to support stored queries at this
 * conformance level.
 * 
 * @see "ISO 19142:2010, cl. A.1.1: Simple WFS"
 * @see "ISO 19142:2010, cl. 11: GetFeature operation"
 * @see "ISO 19142:2010, cl. 7.9.3: Stored query expression"
 */
@Test(groups = "simple")
public class StoredQueryTests extends BaseFixture {

    private Schema wfsSchema;
    private List<String> featureIdentifiers = new ArrayList<String>();

    /**
     * Retrieves the (pre-compiled) WFS schema from the suite fixture.
     * 
     * @param testContext
     *            The test (group) context.
     */
    @BeforeClass
    public void setupClassFixture(ITestContext testContext) {
        this.wfsSchema = (Schema) testContext.getSuite().getAttribute(
                SuiteAttribute.WFS_SCHEMA.getName());
        Assert.assertNotNull(this.wfsSchema,
                "WFS schema not found in suite fixture.");
    }

    /**
     * Builds a DOM Document representing a GetFeature request entity that
     * contains no query expressions.
     */
    @BeforeMethod
    public void buildGetFeatureRequestEntity() {
        this.reqEntity = WFSRequest.createRequestEntity("GetFeature");
    }

    /**
     * [{@code Test}] If no stored query matches the given identifier then an
     * exception report with exception code "InvalidParameterValue" is expected.
     * 
     * @param binding
     *            The ProtocolBinding to use.
     */
    @Test(dataProvider = "protocol-binding")
    public void unknownQueryIdentifier(ProtocolBinding binding) {
        WFSRequest.appendStoredQuery(this.reqEntity,
                "http://docbook.org/ns/docbook",
                Collections.<String, Object> emptyMap());
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        Assert.assertTrue(rsp.hasEntity(),
                ErrorMessage.get(ErrorMessageKeys.MISSING_XML_ENTITY));
        Document entity = extractBodyAsDocument(rsp, binding);
        SchematronValidator validator = ValidationUtils
                .buildSchematronValidator("ExceptionReport.sch",
                        "InvalidParameterValuePhase");
        DOMResult result = validator.validate(new DOMSource(entity));
        Assert.assertFalse(validator.ruleViolationsDetected(), ErrorMessage
                .format(ErrorMessageKeys.NOT_SCHEMA_VALID,
                        validator.getRuleViolationCount(),
                        XMLUtils.writeNodeToString(result.getNode())));
    }

    /**
     * [{@code Test}] Invoking the {@code GetFeatureById} query with an
     * {@code id} parameter value that does not match any feature should produce
     * an empty result set.
     * 
     * @param binding
     *            The ProtocolBinding to use.
     */
    @Test(dataProvider = "protocol-binding")
    public void invokeGetFeatureById_randomID(ProtocolBinding binding) {
        String id = "uuid-" + UUID.randomUUID().toString();
        WFSRequest.appendStoredQuery(this.reqEntity,
                WFS2.QRY_GET_FEATURE_BY_ID,
                Collections.singletonMap("id", (Object) id));
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        Assert.assertTrue(rsp.hasEntity(),
                ErrorMessage.get(ErrorMessageKeys.MISSING_XML_ENTITY));
        Document entity = extractBodyAsDocument(rsp, binding);
        String xpath = "/wfs:FeatureCollection/@numberMatched = 0";
        ETSAssert.assertXPath(xpath, entity.getDocumentElement(), null);
    }

    /**
     * [{@code Test}] Invoking the {@code GetFeatureByType} query should produce
     * a result set containing feature members of the specified type.
     * 
     * @param binding
     *            The ProtocolBinding to use.
     */
    @Test(dataProvider = "protocol-binding")
    public void invokeGetFeatureByType(ProtocolBinding binding) {
        QName featureType = this.featureTypes.get(0);
        WFSRequest.appendStoredQuery(this.reqEntity,
                WFS2.QRY_GET_FEATURE_BY_TYPE,
                Collections.singletonMap("typeName", (Object) featureType));
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.OK.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Assert.assertTrue(rsp.hasEntity(),
                ErrorMessage.get(ErrorMessageKeys.MISSING_XML_ENTITY));
        Document entity = extractBodyAsDocument(rsp, binding);
        ETSAssert.assertQualifiedName(entity.getDocumentElement(), new QName(
                Namespaces.WFS, WFS2.FEATURE_COLLECTION));
        String xpath = "/wfs:FeatureCollection/wfs:member/*";
        Map<String, String> bindings = new HashMap<String, String>();
        bindings.put(Namespaces.WFS, "wfs");
        NodeList members;
        try {
            members = XMLUtils.evaluateXPath(entity, xpath, bindings);
        } catch (XPathExpressionException e) {
            throw new RuntimeException(e);
        }
        if (members.getLength() == 0) {
            // TODO: try another feature type
        }
        Assert.assertTrue(members.getLength() > 0,
                ErrorMessage.format(ErrorMessageKeys.NO_MEMBERS, featureType));
        for (int i = 0; i < members.getLength(); i++) {
            Element feature = (Element) members.item(i);
            String gmlId = feature.getAttributeNS(Namespaces.GML, "id");
            if (!gmlId.isEmpty()) {
                this.featureIdentifiers.add(gmlId);
            }
        }
    }

    /**
     * [{@code Test}] Invoking the {@code GetFeatureById} query with a known
     * feature identifier should produce the matching feature representation in
     * the response.
     * 
     * @param binding
     *            The ProtocolBinding to use.
     */
    @Test(dataProvider = "protocol-binding", dependsOnMethods = { "invokeGetFeatureByType" })
    public void invokeGetFeatureById(ProtocolBinding binding) {
        String featureId = this.featureIdentifiers.get(0);
        WFSRequest.appendStoredQuery(this.reqEntity,
                WFS2.QRY_GET_FEATURE_BY_ID,
                Collections.singletonMap("id", (Object) featureId));
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.GET_FEATURE, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        Assert.assertTrue(rsp.hasEntity(),
                ErrorMessage.get(ErrorMessageKeys.MISSING_XML_ENTITY));
        Document entity = extractBodyAsDocument(rsp, binding);
        ETSAssert.assertQualifiedName(entity.getDocumentElement(), new QName(
                Namespaces.WFS, WFS2.FEATURE_COLLECTION));
        String xpath = "/wfs:FeatureCollection/wfs:member/*";
        Map<String, String> bindings = new HashMap<String, String>();
        bindings.put(Namespaces.WFS, "wfs");
        try {
            NodeList members = XMLUtils.evaluateXPath(entity, xpath, bindings);
            Assert.assertEquals(members.getLength(), 1,
                    "Unexpected number of members in response");
        } catch (XPathExpressionException e) {
            throw new RuntimeException(e);
        }
    }

}
