package org.opengis.cite.iso19142.transaction;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.logging.Level;
import javax.xml.namespace.QName;
import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.BaseFixture;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.cite.iso19142.util.TestSuiteLogger;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Tests the response to a Transaction request that includes one or more delete
 * actions.
 * 
 * @see "ISO 19142:2010, cl. 15.2.7: Delete action"
 */
@Test(groups = "transaction")
public class DeleteTests extends BaseFixture {

    /** List containing original representations of deleted features */
    private List<Element> deletedFeatures = new ArrayList<Element>();

    /**
     * Restores the WFS data store to its previous state by replacing all
     * deleted features with their previous representations.
     */
    @AfterClass
    public void restoreDeletedFeatures() {
        Document rspEntity = this.wfsClient.insert(deletedFeatures);
        String xpath = String.format("//wfs:totalInserted = '%d'",
                deletedFeatures.size());
        ETSAssert.assertXPath(xpath, rspEntity, null);
        TestSuiteLogger.log(
                Level.FINE,
                "Successfully inserted deleted features - "
                        + deletedFeatures.size());
    }

    /**
     * Builds a DOM Document representing a Transaction request entity.
     */
    @BeforeMethod
    public void buildTransactionRequest() {
        this.reqEntity = this.docBuilder.newDocument();
        Element trxElem = this.reqEntity.createElementNS(Namespaces.WFS,
                WFS2.TRANSACTION);
        trxElem.setAttribute(WFS2.SERVICE_PARAM, WFS2.SERVICE_TYPE_CODE);
        trxElem.setAttribute(WFS2.VERSION_PARAM, WFS2.VERSION);
        this.reqEntity.appendChild(trxElem);
    }

    /**
     * Submits a Transaction request to delete an existing feature instance. The
     * test is run for all supported protocol bindings (except HTTP GET) and
     * feature types. The response entity (wfs:TransactionResponse) must be
     * schema-valid and the wfs:TransactionSummary element shall report the
     * correct number of deleted features (totalDeleted).
     * 
     * @param binding
     *            A supported message binding (POST preferred over SOAP).
     * @param featureType
     *            A QName representing the qualified name of some feature type.
     * 
     * @see "ISO 19142:2010, cl. 15.3.3: TransactionSummary element"
     */
    @Test(dataProvider = "protocol-featureType")
    public void deleteFeature(ProtocolBinding binding, QName featureType) {
        Document doc = wfsClient.getFeatureByType(featureType, 10);
        NodeList features = doc.getElementsByTagNameNS(
                featureType.getNamespaceURI(), featureType.getLocalPart());
        // randomly select an available feature to delete
        Random random = new Random();
        Element originalFeature = (Element) features.item(random
                .nextInt(features.getLength()));
        String gmlId = originalFeature.getAttributeNS(Namespaces.GML, "id");
        Map<String, QName> featuresToDelete = new HashMap<String, QName>();
        featuresToDelete.put(
                gmlId,
                new QName(originalFeature.getNamespaceURI(), originalFeature
                        .getLocalName()));
        Document rspEntity = wfsClient.delete(featuresToDelete);
        ETSAssert.assertXPath("//wfs:TransactionResponse", rspEntity, null);
        String xpath = String.format("//wfs:totalDeleted = '%d'",
                featuresToDelete.size());
        ETSAssert.assertXPath(xpath, rspEntity, null);
        ETSAssert.assertFeatureAvailability(gmlId, false, wfsClient);
        deletedFeatures.add(originalFeature);
    }
}
