package org.opengis.cite.iso19142.transaction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import javax.xml.namespace.QName;
import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.BaseFixture;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.cite.iso19142.util.XMLUtils;
import org.opengis.cite.iso19142.util.TestSuiteLogger;
import org.opengis.cite.iso19142.util.WFSRequest;
import org.testng.Assert;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Tests the response to a Transaction request that includes one or more replace
 * actions.
 * 
 * @see "ISO 19142:2010, cl. 15.2.6: Replace action"
 */
@Test(groups = "transaction")
public class ReplaceTests extends BaseFixture {

    /** List containing original representations of modified features */
    private List<Element> modifiedFeatures = new ArrayList<Element>();

    /**
     * Restores the WFS data store to its previous state by replacing all
     * modified features with their original representations.
     */
    @AfterClass
    public void restoreModifiedFeatures() {
        Document rspEntity = this.wfsClient.replace(modifiedFeatures);
        String xpath = String.format("//wfs:totalReplaced = '%d'",
                modifiedFeatures.size());
        ETSAssert.assertXPath(xpath, rspEntity, null);
        TestSuiteLogger.log(Level.FINE, "Replaced features - "
                + modifiedFeatures.size());
    }

    /**
     * Builds a DOM Document representing a Transaction request entity.
     */
    @BeforeMethod
    public void buildTransactionRequest() {
        this.reqEntity = this.docBuilder.newDocument();
        Element trxElem = this.reqEntity.createElementNS(Namespaces.WFS,
                WFS2.TRANSACTION);
        trxElem.setAttribute(WFS2.SERVICE_PARAM, WFS2.SERVICE_TYPE_CODE);
        trxElem.setAttribute(WFS2.VERSION_PARAM, WFS2.VERSION);
        this.reqEntity.appendChild(trxElem);
    }

    /**
     * Submits a Transaction request to replace an existing feature instance.
     * The test is run for all supported protocol bindings (except HTTP GET) and
     * feature types. The response entity (wfs:TransactionResponse) must be
     * schema-valid and contain the wfs:ReplaceResults element.
     * 
     * @param binding
     *            A supported message binding.
     * @param featureType
     *            A QName representing the qualified name of some feature type.
     * 
     * @see "ISO 19142:2010, cl. 15.3.3: TransactionSummary element"
     * @see "ISO 19142:2010, cl. 15.3.6: ReplaceResults element"
     */
    @Test(dataProvider = "protocol-featureType")
    public void replaceFeature(ProtocolBinding binding, QName featureType) {
        Document doc = wfsClient.getFeatureByType(featureType, 1);
        NodeList features = doc.getElementsByTagNameNS(
                featureType.getNamespaceURI(), featureType.getLocalPart());
        Element originalFeature = (Element) features.item(0);
        Element replacement = createReplacementFeature(originalFeature);
        List<Element> replacements = Arrays.asList(replacement);
        Document rspEntity = wfsClient.replace(replacements);
        ETSAssert.assertXPath("//wfs:TransactionResponse", rspEntity, null);
        String xpath = String.format("//wfs:totalReplaced = '%d'",
                replacements.size());
        ETSAssert.assertXPath(xpath, rspEntity, null);
        Node updateResults = rspEntity.getElementsByTagNameNS(Namespaces.WFS,
                "ReplaceResults").item(0);
        Assert.assertNotNull(updateResults,
                "Expected wfs:ReplaceResults element in response.");
        modifiedFeatures.add(originalFeature);
    }

    /**
     * Creates a new feature representation by cloning and modifying the
     * original instance. The following standard GML properties are added (or
     * replaced if already present):
     * 
     * <ul>
     * <li>gml:description ("Lorem ipsum dolor sit amet.")</li>
     * <li>gml:identifier (random UUID value)</li>
     * </ul>
     * 
     * @param originalFeature
     *            The original feature instance (unmodified).
     * @return A new Element node representing the modified feature.
     */
    Element createReplacementFeature(Element originalFeature) {
        Element replacement = (Element) originalFeature.cloneNode(true);
        QName propName = new QName(Namespaces.GML, "identifier");
        Element identifier = XMLUtils.createElement(propName);
        identifier.setAttribute("codeSpace", "http://cite.opengeospatial.org/");
        identifier.setTextContent(UUID.randomUUID().toString());
        WFSRequest.insertGMLProperty(replacement, identifier);
        propName = new QName(Namespaces.GML, "description");
        Element desc = XMLUtils.createElement(propName);
        desc.setTextContent("Lorem ipsum dolor sit amet.");
        WFSRequest.insertGMLProperty(replacement, desc);
        return replacement;
    }
}
