package org.opengis.cite.iso19142.transaction;

import com.sun.jersey.api.client.ClientResponse;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;
import org.opengis.cite.iso19142.ETSAssert;
import org.opengis.cite.iso19142.ErrorMessage;
import org.opengis.cite.iso19142.ErrorMessageKeys;
import org.opengis.cite.iso19142.BaseFixture;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.cite.iso19142.util.ServiceMetadataUtils;
import org.opengis.cite.iso19142.util.TestSuiteLogger;
import org.opengis.cite.iso19142.util.WFSRequest;
import org.testng.Assert;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Tests the response to a Transaction request that includes one or more update
 * actions.
 * 
 * @see "ISO 19142:2010, cl. 15.2.5: Update action"
 */
@Test(groups = "transaction")
public class UpdateTests extends BaseFixture {

    /** List containing original representations of modified features */
    private List<Element> modifiedFeatures = new ArrayList<Element>();

    /**
     * Restores the WFS data store to its previous state by replacing all
     * modified features with their original representations.
     */
    @AfterClass
    public void replaceUpdatedFeatures() {
        Document rspEntity = this.wfsClient.replace(modifiedFeatures);
        String xpath = String.format("//wfs:totalReplaced = '%d'",
                modifiedFeatures.size());
        ETSAssert.assertXPath(xpath, rspEntity, null);
        TestSuiteLogger.log(Level.FINE, "Replaced features - "
                + modifiedFeatures.size());
    }

    /**
     * Builds a DOM Document representing a Transaction request entity.
     */
    @BeforeMethod
    public void buildTransactionRequest() {
        this.reqEntity = this.docBuilder.newDocument();
        Element trxElem = this.reqEntity.createElementNS(Namespaces.WFS,
                WFS2.TRANSACTION);
        trxElem.setAttribute(WFS2.SERVICE_PARAM, WFS2.SERVICE_TYPE_CODE);
        trxElem.setAttribute(WFS2.VERSION_PARAM, WFS2.VERSION);
        this.reqEntity.appendChild(trxElem);
    }

    /**
     * Submits a Transaction request to update the name (gml:name[1]) of an
     * existing feature instance. The test is run for all supported protocol
     * bindings (except HTTP GET) and feature types. The response entity
     * (wfs:TransactionResponse) must be schema-valid and contain the
     * wfs:UpdateResults element.
     * 
     * @param binding
     *            A supported message binding.
     * @param featureType
     *            A QName representing the qualified name of some feature type.
     * 
     * @see "ISO 19142:2010, cl. 15.3.3: TransactionSummary element"
     * @see "ISO 19142:2010, cl. 15.3.5: UpdateResults element"
     */
    @Test(dataProvider = "protocol-featureType")
    public void updateFeatureName(ProtocolBinding binding, QName featureType) {
        Document doc = wfsClient.getFeatureByType(featureType, 1);
        NodeList features = doc.getElementsByTagNameNS(
                featureType.getNamespaceURI(), featureType.getLocalPart());
        Element originalFeature = (Element) features.item(0);
        String gmlId = originalFeature.getAttributeNS(Namespaces.GML, "id");
        Map<String, Object> properties = new HashMap<String, Object>();
        properties.put("gml:name[1]", "Ce n'est pas Montréal!");
        properties.put("gml:name[1]/@codeSpace",
                "http://cite.opengeospatial.org/");
        Document rspEntity = wfsClient.updateFeature(gmlId, featureType,
                properties);
        ETSAssert.assertXPath("//wfs:TransactionResponse", rspEntity, null);
        Node updateResults = rspEntity.getElementsByTagNameNS(Namespaces.WFS,
                "UpdateResults").item(0);
        Assert.assertNotNull(updateResults,
                "Expected wfs:UpdateResults element in response.");
        ETSAssert.assertFeatureProperties(gmlId, properties, null, wfsClient);
        modifiedFeatures.add(originalFeature);
    }

    /**
     * Submits a Transaction request to update a property (gml:boundedBy) with
     * an invalid value (kml:Point). An ExceptionReport (with status code 400)
     * containing the exception code {@code InvalidValue} is expected in
     * response.
     * 
     * @see "ISO 19142:2010, Table 3:  WFS exception codes"
     */
    @Test
    public void updateInvalidFeatureProperty() {
        try {
            this.reqEntity = docBuilder.parse(getClass().getResourceAsStream(
                    "UpdateInvalidFeatureProperty.xml"));
        } catch (Exception e) {
            throw new RuntimeException(
                    "Failed to parse XML resource from classpath", e);
        }
        Element update = (Element) this.reqEntity.getElementsByTagNameNS(
                Namespaces.WFS, WFS2.UPDATE).item(0);
        WFSRequest.setTypeName(update, featureTypes.get(0));
        ProtocolBinding binding = wfsClient.getTransactionBinding();
        URI endpoint = ServiceMetadataUtils.getOperationEndpoint(
                this.wfsMetadata, WFS2.TRANSACTION, binding);
        ClientResponse rsp = wfsClient.submitRequest(new DOMSource(
                this.reqEntity), binding, endpoint);
        Assert.assertEquals(rsp.getStatus(),
                ClientResponse.Status.BAD_REQUEST.getStatusCode(),
                ErrorMessage.get(ErrorMessageKeys.UNEXPECTED_STATUS));
        Document entity = rsp.getEntity(Document.class);
        String xpath = "//ows:Exception[@exceptionCode = 'InvalidValue']";
        ETSAssert.assertXPath(xpath, entity.getDocumentElement(), null);
    }
}
