package org.opengis.cite.iso19142.util;

import java.net.URI;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import javax.xml.namespace.QName;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.opengis.cite.iso19142.FeatureTypeInfo;
import org.opengis.cite.iso19142.Namespaces;
import org.opengis.cite.iso19142.ProtocolBinding;
import org.opengis.cite.iso19142.WFS2;
import org.opengis.util.FactoryException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Provides various utility methods for accessing service metadata.
 */
public class ServiceMetadataUtils {

    /**
     * Extracts a request endpoint from a WFS capabilities document. If the
     * request URI contains a query component it is removed (but not from the
     * source document).
     * 
     * @param wfsMetadata
     *            A DOM Document node containing service metadata (OGC
     *            capabilities document).
     * @param opName
     *            The operation (request) name.
     * @param binding
     *            The MessageBinding to use.
     * @return A URI referring to a request endpoint; the URI is empty if no
     *         matching endpoint is found.
     */
    public static URI getOperationEndpoint(final Document wfsMetadata,
            String opName, ProtocolBinding binding) {
        if (binding.equals(ProtocolBinding.SOAP)) {
            // use POST method for SOAP request
            binding = ProtocolBinding.POST;
        }
        // method name in OGC capabilities doc has initial capital
        StringBuilder method = new StringBuilder(binding.toString());
        method.replace(1, method.length(), method.substring(1).toLowerCase());
        NamespaceBindings nsBindings = new NamespaceBindings();
        nsBindings.addNamespaceBinding(Namespaces.OWS, "ows");
        nsBindings.addNamespaceBinding(Namespaces.XLINK, "xlink");
        String expr = String.format(
                "//ows:Operation[@name='%s']//ows:%s/@xlink:href", opName,
                method.toString());
        XPath xpath = XPathFactory.newInstance().newXPath();
        xpath.setNamespaceContext(nsBindings);
        URI endpoint = null;
        try {
            String href = xpath.evaluate(expr, wfsMetadata);
            endpoint = URI.create(href);
        } catch (XPathExpressionException ex) {
            // XPath expression is correct
            TestSuiteLogger.log(Level.INFO, ex.getMessage());
        }
        if (null != endpoint.getQuery()) {
            // prune query component if present
            String uri = endpoint.toString();
            endpoint = URI.create(uri.substring(0, uri.indexOf('?')));
        }
        return endpoint;
    }

    /**
     * Returns a Map containing the HTTP endpoints for a given service request.
     * 
     * @param wfsMetadata
     *            A DOM Document node containing service metadata (WFS
     *            capabilities document).
     * @param reqName
     *            The (local) name of the service request.
     * @return A Map<String, URI> object that associates an HTTP method name
     *         with a URI, or {@code null} if the request is not implemented.
     */
    public static Map<String, URI> getRequestEndpoints(
            final Document wfsMetadata, String reqName) {
        NamespaceBindings nsBindings = NamespaceBindings.withStandardBindings();
        String expr = String.format(
                "//ows:Operation[@name='%s']/descendant::*[@xlink:href]",
                reqName);
        XPath xpath = XPathFactory.newInstance().newXPath();
        xpath.setNamespaceContext(nsBindings);
        Map<String, URI> endpoints = null;
        try {
            NodeList methodNodes = (NodeList) xpath.evaluate(expr, wfsMetadata,
                    XPathConstants.NODESET);
            if ((null == methodNodes) || methodNodes.getLength() == 0) {
                return null;
            }
            endpoints = new HashMap<String, URI>();
            for (int i = 0; i < methodNodes.getLength(); i++) {
                Element methodElem = (Element) methodNodes.item(i);
                String methodName = methodElem.getLocalName().toUpperCase();
                String href = methodElem.getAttributeNS(Namespaces.XLINK,
                        "href");
                if (href.indexOf('?') > 0) {
                    // prune query component if present
                    href = href.substring(0, href.indexOf('?'));
                }
                endpoints.put(methodName, URI.create(href));
            }
        } catch (XPathExpressionException xpe) {
            TestSuiteLogger.log(Level.INFO, xpe.getMessage());
        }
        return endpoints;
    }

    /**
     * Extracts the list of feature type names from a WFS capabilities document.
     * 
     * @param wfsMetadata
     *            A service capabilities document (wfs:WFS_Capabilities).
     * @return A List containing one or more QName items.
     */
    public static List<QName> getFeatureTypes(final Document wfsMetadata) {
        String xpath = "//wfs:FeatureType/wfs:Name";
        Map<String, String> nsBindings = new HashMap<String, String>();
        nsBindings.put(Namespaces.WFS, "wfs");
        NodeList typeNames = null;
        try {
            typeNames = XMLUtils.evaluateXPath(wfsMetadata, xpath, nsBindings);
        } catch (XPathExpressionException xpe) {
            TestSuiteLogger.log(Level.INFO,
                    "Failed to evaluate XPath expression: " + xpath, xpe);
        }
        List<QName> featureTypes = new ArrayList<QName>();
        for (int i = 0; i < typeNames.getLength(); i++) {
            Node typeName = typeNames.item(i);
            featureTypes.add(buildQName(typeName));
        }
        return featureTypes;
    }

    /**
     * Extracts information about feature types from the service metadata
     * document. The following information items are collected for each feature
     * type:
     * 
     * <ul>
     * <li>Qualified type name (QName)</li>
     * <li>Default CRS reference</li>
     * <li>Spatial extent</li>
     * </ul>
     * 
     * @param wfsCapabilities
     *            A Document (wfs:WFS_Capabilities).
     * @return A Map containing one or more entries where a feature type name
     *         (QName) is associated with a FeatureTypeInfo value object.
     */
    public static Map<QName, FeatureTypeInfo> extractFeatureInfo(
            final Document wfsCapabilities) {
        Map<QName, FeatureTypeInfo> featureInfo = new HashMap<QName, FeatureTypeInfo>();
        NodeList featureTypes = wfsCapabilities.getElementsByTagNameNS(
                Namespaces.WFS, "FeatureType");
        for (int i = 0; i < featureTypes.getLength(); i++) {
            FeatureTypeInfo typeInfo = new FeatureTypeInfo();
            NodeList children = featureTypes.item(i).getChildNodes();
            for (int j = 0; j < children.getLength(); j++) {
                Node child = children.item(j);
                if (child.getNodeType() != Node.ELEMENT_NODE) {
                    continue;
                }
                if (child.getLocalName().equals("Name")) {
                    QName typeName = buildQName(child);
                    typeInfo.setTypeName(typeName);
                }
                if (child.getLocalName().equals("DefaultCRS")) {
                    try {
                        typeInfo.setDefaultCRS(child.getTextContent());
                    } catch (FactoryException e) {
                        TestSuiteLogger.log(Level.WARNING,
                                "Failed to set default CRS.", e);
                    }
                }
                // Ignore ows:WGS84BoundingBox as it is rarely accurate
            }
            featureInfo.put(typeInfo.getTypeName(), typeInfo);
        }
        return featureInfo;
    }

    /**
     * Builds a QName representing the qualified name conveyed by a node with
     * simple content.
     * 
     * @param node
     *            A DOM node (Element) containing a qualified name (xsd:QName)
     *            as text content.
     * @return A QName object.
     */
    public static QName buildQName(Node node) {
        String[] parts = node.getTextContent().split(":");
        String nsName = node.lookupNamespaceURI(parts[0]);
        return new QName(nsName, parts[1]);
    }

    /**
     * Discovers which protocol bindings are implemented by a WFS. The values of
     * these standard service constraints are checked:
     * 
     * <ul>
     * <li>KVPEncoding</li>
     * <li>XMLEncoding</li>
     * <li>SOAPEncoding</li>
     * </ul>
     * 
     * @param wfsMetadata
     *            A service metadata document (wfs:WFS_Capabilities).
     * @return A Set of protocol bindings implemented by the SUT.
     */
    public static Set<ProtocolBinding> getSupportedBindings(
            final Document wfsMetadata) {
        if (null == wfsMetadata) {
            throw new NullPointerException("WFS metadata document is null.");
        }
        Set<ProtocolBinding> supportedBindings = EnumSet
                .noneOf(ProtocolBinding.class);
        String xpath = "//ows:Constraint[@name='%s' and (ows:DefaultValue = 'TRUE')]";
        Map<String, String> nsBindings = new HashMap<String, String>();
        nsBindings.put(Namespaces.OWS, "ows");
        try {
            if (XMLUtils.evaluateXPath(wfsMetadata,
                    String.format(xpath, WFS2.HTTP_GET), nsBindings)
                    .getLength() > 0) {
                supportedBindings.add(ProtocolBinding.GET);
            }
            if (XMLUtils.evaluateXPath(wfsMetadata,
                    String.format(xpath, WFS2.HTTP_POST), nsBindings)
                    .getLength() > 0) {
                supportedBindings.add(ProtocolBinding.POST);
            }
            if (XMLUtils.evaluateXPath(wfsMetadata,
                    String.format(xpath, WFS2.SOAP), nsBindings).getLength() > 0) {
                supportedBindings.add(ProtocolBinding.SOAP);
            }
        } catch (XPathExpressionException xpe) {
            throw new RuntimeException(
                    "Error evaluating XPath expression against capabilities doc. ",
                    xpe);
        }
        return supportedBindings;
    }
}
